/*
 * fifo.h
 * 
 * Copyright (C) 2005  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * 
 *****************************************************************************
 *
 * pozadavky:
 * 	- jeden odesilatel (funguje vice)
 * 	- vice prijemcu
 * 	- zarucit poradi
 *
 * 
 */
#ifndef JK_FIFO_H
#define JK_FIFO_H

#include <stdio.h>
#include <string.h>
#include <pthread.h>
#include <queue>

#include "ref.h"

#define CLASS "Fifo"
#ifdef DEBUG
#	define FIFO_H__DEBUG DEBUG
#	undef DEBUG
#endif
#define DEBUG 0
#define DEBUG_PREFIX_STR	"%s: "
#define DEBUG_PREFIX_ARG	name
#include "debug.h"

template <class T>
class Fifo {
	public:
		Fifo(char *name);
		~Fifo(void);

		// destroy fifo -- never call from signal handler in same
		// thread as pop(), push()
		void destroy(void);

		// Add to end of fifo
		bool push(T obj);

		// get from begin of fifo
		bool pop(T *obj);

		// return true if fifo is empty
		inline bool is_hungry();

		// (un)register source
		int source_ref();
		int source_unref();

		// (un)register target
		int target_ref();
		int target_unref();

	private:
		bool alive;			// je fifo v provozu ?
		char *name;
		int use_count;			// pocet prave pristupujicich
		pthread_cond_t destroy_cond;	// cekani na zniceni
		
		unsigned int targets_count,	// pocet prijemcu
			     sources_count;	// pocet odesilatelu
		
		pthread_mutex_t access_lock,	// zamek pristupu k cemukoliv
				use_lock;	// zamek pro pocitani pristupu
		
		pthread_cond_t wait_cond,	// cekani na ostatni
			       empty_cond,	// cekani pri prazdne fifo
			       full_cond;	// cekani pri plne fifo

		// hloupa kruhova fronta
		class CircBuf {
			public:
				CircBuf(int size) {
					buffer = new T[size];
					bufsize = size;
					head = tail = 0;
				}
				~CircBuf(void) {
					delete[] buffer;
				}
				
				void push(T obj) {
					buffer[head] = obj;
					head = (head + 1) % bufsize;
				}
				
				T front(void) {
					return buffer[tail];
				}
				
				void pop(void) {
					tail = (tail + 1) % bufsize;
				}

				bool is_full(void) {
					return (tail + 1) % bufsize == head;
				}

				bool is_empty(void) {
					return tail == head;
				}

				int size(void) {
					return bufsize;
				}
			private:
				T *buffer;
				int head, tail, bufsize;
		};

		CircBuf *fifo;
		
		// data
		static const unsigned int fifo_max_size;
		unsigned int pop_wait_counter;
		//std::queue<T> fifo;

		// pocitani prave pristupujicich vlaken
		inline void use_ref(void);
		inline void use_unref(void);

};


/****************************************************************************/

/*
 * g++-3.3 -Wall -D_REENTRANT -DSOUND_OSS -DFORCE_DEBUG=-1  -c fifo.cc -o fifo.o
 * fifo.cc:26: warning: keyword `export' not implemented, and will be ignored
 */

/****************************************************************************/

template <class T>
const unsigned int Fifo<T>::fifo_max_size = 10;
		
template <class T>
Fifo<T>::Fifo(char *name) {
	if (!name) {
		this->name = strdup("unnamed");
	} else {
		this->name = strdup(name);
	}
	DBG_MSG(8, "creating...");
	alive = true;
	use_count = 0;

	fifo = new CircBuf(fifo_max_size);

	pthread_mutex_init(&access_lock, NULL);
	pthread_cond_init(&full_cond, NULL);
	pthread_cond_init(&empty_cond, NULL);
	pthread_cond_init(&wait_cond, NULL);
	pthread_cond_init(&destroy_cond, NULL);
	
	targets_count = 0;
	sources_count = 1;
	pop_wait_counter = 0;

	DBG_MSG(4, "Fifo \"%s\" done.", this->name);
}

/****************************************************************************/

template <class T>
Fifo<T>::~Fifo(void) {
	DBG_MSG(8, "removing...");
	destroy();
	free(name);
}

/****************************************************************************/

template <class T>
void Fifo<T>::destroy(void) {	
	if(!alive)
		return;
	DBG_MSG(7, "");
	pthread_mutex_lock(&use_lock);
	alive = false;
	while (use_count > 0) {
		DBG_MSG(7, "Canceling running operations... %d remains.", use_count);
		pthread_cond_broadcast(&full_cond);
		pthread_cond_broadcast(&empty_cond);
		pthread_cond_broadcast(&wait_cond);
		pthread_cond_wait(&destroy_cond, &use_lock);
	}
	pthread_mutex_unlock(&use_lock);

	DBG_MSG(7, "Destroying fifo...");
	pthread_mutex_destroy(&access_lock);
	pthread_cond_destroy(&full_cond);
	pthread_cond_destroy(&empty_cond);
	pthread_cond_destroy(&wait_cond);
	pthread_cond_destroy(&destroy_cond);

	delete fifo;

	DBG_MSG(7, "Done.");
}

/****************************************************************************/

template <class T>
bool Fifo<T>::push(T obj) {
	if(!alive)
		return false;
	use_ref();
	DBG_MSG(7, "access_lock...");
	pthread_mutex_lock(&access_lock);
	DBG_MSG(7, "access_lock = on");
	
	while (fifo->is_full()) {
		if (!alive) {
			DBG_MSG(6, "Fifo is dead.");
			DBG_MSG(7, "access_lock = off");
			pthread_mutex_unlock(&access_lock);
			use_unref();
			return false;
		}
		DBG_MSG(6, "Fifo is full, waiting...");
		pthread_cond_wait(&full_cond, &access_lock);
	}
	
	if (obj)
		obj->ref();
	
	fifo->push(obj);
	
	DBG_MSG(8, "objects left in fifo: %d", fifo->size());
	pthread_cond_broadcast(&empty_cond);
	
	DBG_MSG(7, "access_lock = off");
	pthread_mutex_unlock(&access_lock);
	use_unref();
	return true;
}

/****************************************************************************/

template <class T>
bool Fifo<T>::pop(T *obj) {
	if(!alive)
		return false;
	use_ref();
	DBG_MSG(7, "access_lock...");
	pthread_mutex_lock(&access_lock);
	DBG_MSG(7, "access_lock = on");
	
	while (fifo->is_empty()) {
		if (!alive) {
			DBG_MSG(6, "Fifo is dead.");
			DBG_MSG(7, "access_lock = off");
			pthread_mutex_unlock(&access_lock);
			use_unref();
			return false;
		}

		if (sources_count == 0 && fifo->is_empty()) {
			DBG_MSG(6, "Fifo is hungry.");
			DBG_MSG(7, "access_lock = off");
			pthread_mutex_unlock(&access_lock);
			use_unref();
			return false;
		}

		DBG_MSG(6, "Fifo is empty, waiting...");
		pthread_cond_wait(&empty_cond, &access_lock);
	}
	
	// precteni dat
	DBG_MSG(6, "Reading data");
	*obj = fifo->front();
	if (*obj)
		(*obj)->ref();
	DBG_MSG(8, "objects left in fifo: %d", fifo->size());

	if (++pop_wait_counter >= targets_count) {
		// jsme posledni -- likvidace objektu + uvolneni ostatnich
		DBG_MSG(6, "I'm last -- removing data");
		if(*obj)
			(*obj)->unref();
		fifo->pop();
		pthread_cond_broadcast(&full_cond);

		pop_wait_counter = 0;
		pthread_cond_broadcast(&wait_cond);
	} else {
		// pockame na ostatni
		DBG_MSG(6, "Waiting for others");
		while (pthread_cond_wait(&wait_cond, &access_lock) != 0) {
			if (!alive) {
				DBG_MSG(6, "Fifo is dead.");
				DBG_MSG(7, "access_lock = off");
				pthread_mutex_unlock(&access_lock);
				use_unref();
				return false;
			}
		}
		DBG_MSG(8, "Others are ready");
	}

	DBG_MSG(7, "access_lock = off");
	pthread_mutex_unlock(&access_lock);
	use_unref();
	return true;
}

/****************************************************************************/

template <class T>
bool Fifo<T>::is_hungry() {
	if(!alive)
		return false;
	return (sources_count == 0 && fifo.empty());
}

/****************************************************************************/

template <class T>
int Fifo<T>::source_ref() {
	DBG_MSG(8, "");
	if(!alive)
		return -1;
	return ++sources_count;
}

/****************************************************************************/

template <class T>
int Fifo<T>::source_unref() {
	DBG_MSG(8, "");
	if (sources_count) {
		sources_count--;
		pthread_cond_broadcast(&empty_cond);
		DBG_MSG(6, "Source unref");
		return sources_count;
	} else {
		DBG_MSG(1, "CRITICAL: sources_count == 0 !");
		return 0;
	}
}

/****************************************************************************/

template <class T>
int Fifo<T>::target_ref() {
	DBG_MSG(8, "");
	if(!alive)
		return -1;
	return ++targets_count;
}

/****************************************************************************/

template <class T>
int Fifo<T>::target_unref() {
	DBG_MSG(8, "");
	if (targets_count)
		return --targets_count;
	else {
		DBG_MSG(1, "CRITICAL: targets_count == 0 !");
		return 0;
	}
}

/****************************************************************************/

template <class T>
void Fifo<T>::use_ref(void) {
	pthread_mutex_lock(&use_lock);
	use_count++;
	pthread_mutex_unlock(&use_lock);
}

template <class T>
void Fifo<T>::use_unref(void) {
	pthread_mutex_lock(&use_lock);
	use_count--;
	if (use_count <= 0) {
		DBG_MSG(8, "Nothing.");
	}
	pthread_cond_broadcast(&destroy_cond);
	pthread_mutex_unlock(&use_lock);
}


#undef DEBUG
#ifdef FIFO_H__DEBUG
#	define DEBUG FIFO_H__DEBUG
#	undef FIFO_H__DEBUG
#endif

#undef DEBUG_PREFIX_STR
#undef DEBUG_PREFIX_ARG
#undef CLASS
#endif

