/*
 * game.cc
 * 
 * Copyright (C) 2005  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>

#include "game.h"
#include "game_object.h"
#include "game_world.h"
#include "sample_msg.h"
#include "screen.h"
#include "screen_msg.h"
#include "sound_const.h"

#define CLASS "Game"
#define DEBUG 0
#include "debug.h"

const Game::KeyBound Game::key_bound[KEY_NUM] = {
	{    3,    6, 'k' },	// krtek
	{    6,   10, 'K' },
	{   24,   34, 'z' },	// zahradnik
	{   36,   47, 'Z' }
};

Game::Game(SampleFifo *in_fifo, ScreenFifo *out_fifo, int height, int width) {
	DBG_MSG(1, "Initializing...");
	
	this->main_done = 0;
	this->world_height = height;
	this->world_width = width;
	this->in_fifo = in_fifo;
	this->out_fifo = out_fifo;
	// this->objects;
	
	in_fifo->target_ref();
	out_fifo->source_ref();
	
	DBG_MSG(2, "Done.");
}

Game::~Game(void) {
	DBG_MSG(1, "Cleaning up...");
	
	main_done = 1;
	
	DBG_MSG(2, "Done.");
}

void Game::main(void) {
	SampleMsg *msg;
	ScreenMsg *scr_msg;
	GameWorld *world;
	GObject *object;
	int time = 0;
	int key[KEY_NUM];
	int old_key[KEY_NUM];		// stare klavesy

	DBG_MSG(1, "Started.");

	world = new GameWorld(world_width, world_height);

	/* define freq rule and world size */
	ScreenFreqRule *rule = new ScreenFreqRule();
	for (int b = 0; b < KEY_NUM; b++) {
		rule->add(key_bound[b].min, key_bound[b].max, key_bound[b].symbol);
	}
	scr_msg = new ScreenMsg();
	scr_msg->add_freq_rule(rule);
	scr_msg->set_world_size(world_width, world_height);
	out_fifo->push(scr_msg);
	scr_msg->unref();
	
	DBG_MSG(2, "Starting main loop...");
	while (!main_done) {
		DBG_MSG(5, "Waiting for event...");
		if (!in_fifo->pop(&msg)) {
			break;
		}
		if (msg == NULL) {
			DBG_MSG(5, "Timeout received.");
			/* timeout */
			time++;
			world->step();
			scr_msg = new ScreenMsg();

			// set world size
			scr_msg->set_world_size(world->get_width(), world->get_height());

			// draw all
			world->get_child_rewind();
			while ((object = world->get_child())) {
				scr_msg->add_symbol(object->draw());
			}

			if (!out_fifo->push(scr_msg))
				main_done = 1;
			scr_msg->unref();

		} else {
			DBG_MSG(5, "SampleMsg received.");
			
			/* sample message */
			sample2keys(key, msg, 1);
			
			// forward sample message to screen, send status
			scr_msg = new ScreenMsg();
			scr_msg->add_sample_msg(msg);
			scr_msg->set_status(key[0] | (key[1] << 1), world->get_score_krtek(),
					key[2] | (key[3] << 1), world->get_score_zahradnik());
			if (!out_fifo->push(scr_msg))
				main_done = 1;
			scr_msg->unref();


			/* odstraneni dlouhych stisku */
			for (int k = 0; k < KEY_NUM; k++) {
				if (key[k]) {
					if (old_key[k])
						key[k] = 0;
					old_key[k] = 1;
				} else {
					old_key[k] = 0;
				}
			}
			world->interpret_keys(key);
			msg->unref();
			//screen->flush();
		}
	}

	in_fifo->target_unref();
	out_fifo->source_unref();

	DBG_MSG(1, "Exit.");
	return;
}

void Game::kill(void) {
	DBG_MSG(1, "Hey Game::main, you are dead!");
	main_done = 1;
}

/*****************************************************************************
 * Mstn metody
 ****************************************************************************/

/*****************************************************************************
 * Prevod frekvencniho spektra na klavesy (akce)
 *
 * Detekuje frekvence s velkou amplitudou a podle toho nastavuje flagy
 * jednotlivych pasem. Na zacatku je male pasmo, ktere se vynechava (nejnizsi
 * frekvence = zadny zvuk).
 *
 */
void Game::sample2keys(int key[KEY_NUM], SampleMsg *msg, int clear) {
	
	if (clear)
		bzero(key, sizeof(int) * KEY_NUM);

	/* pokud neni signal dostatecne silny, je na nic */
	if (msg->intensity < MIN_INTENSITY)
		return;
	
	/* scan spektra */
	for (int i = 0; i < msg->size && msg->data[i].val > msg->freq_minval; i++) {
		for (int b = 0; b < KEY_NUM; b++) {
			if (msg->data[i].freq > key_bound[b].min && msg->data[i].freq < key_bound[b].max) {
				key[b] = 1;
				break;
			}
		}
	}
}

#undef CLASS

