/*
 * game_glmodel.cc
 * 
 * Copyright (C) 2006  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <GL/glut.h>
#include <GL/glpng.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <strings.h>
#include "game_glmodel.h"

#define DEBUG 5
#define CLASS "GGLModel"
#include "debug.h"

#define MAX_TYPE_LEN	64

GGLModel::GGLModel(void) {
}

GGLModel::~GGLModel(void) {
}

bool GGLModel::load(char *filename) {
	FILE *file;
	bool first_vertex = true;
	MATERIAL *usemtl = NULL;
	
	if (!filename)
		return false;

	// open file
	if (!(file = fopen(filename, "rt"))) {
		perror(filename);
		return false;
	}

	// prepare data storage
	geom_vertex.clear();
	norm_vertex.clear();
	texture_vertex.clear();
	material.clear();
	face.clear();

	// read & parse all lines
	while (!feof(file)) {
		char *word = NULL;
		if (fscanf(file, " %as", &word) == 1) {
			if (word[0] == '#') {
				// comment
				//DBG_MSG(6, "Skipped comment");
				
			} else if (strcmp(word, "v") == 0) {
				// geometry vertex
				VERTEX v;
				fscanf(file, " %f %f %f", &v.x, &v.y, &v.z);
				geom_vertex.push_back(v);
				DBG_MSG(6, "Loaded geometry vertex: %.3f, %.3f, %.3f", v.x, v.y, v.z);
				
				size_min[0] = first_vertex || size_min[0] > v.x ? v.x : size_min[0];
				size_min[1] = first_vertex || size_min[1] > v.y ? v.y : size_min[1];
				size_min[2] = first_vertex || size_min[2] > v.z ? v.z : size_min[2];
				size_max[0] = first_vertex || size_max[0] < v.x ? v.x : size_max[0];
				size_max[1] = first_vertex || size_max[1] < v.y ? v.y : size_max[1];
				size_max[2] = first_vertex || size_max[2] < v.z ? v.z : size_max[2];
				first_vertex = false;
					
			} else if (strcmp(word, "vn") == 0) {
				// normal vertex
				VERTEX v;
				fscanf(file, " %f %f %f", &v.x, &v.y, &v.z);
				norm_vertex.push_back(v);
				DBG_MSG(6, "Loaded normal vertex: %.3f, %.3f, %.3f", v.x, v.y, v.z);
					
			} else if (strcmp(word, "vt") == 0) {
				// texture vertex
				VERTEX v;
				fscanf(file, " %f %f", &v.x, &v.y);
				v.z = 0.0;
				texture_vertex.push_back(v);
				DBG_MSG(6, "Loaded texture vertex: %.3f, %.3f", v.x, v.y, v.z);
					
			} else if (strcmp(word, "f") == 0) {
				// face
				FACE f;
				FACE_VERTEX fv;
				char *str;
				
				DBG_MSG(6, "Face");
				fv.g = -1;
				fv.t = -1;
				fv.n = -1;
				fv.material = usemtl;
				while (skip_space(file) && fscanf(file, "%a[0-9/-]s", &str) == 1) {
					int *val[] = { &fv.g, &fv.t, &fv.n };
					int i = 0;
					char *p = str;
					while (i < 3 && p) {
						if (i > 0)
							p++;
						if (sscanf(p, "%d", val[i]) != 1)
							*val[i] = -1;
						p = strchr(p, '/');
						i++;
					}
						
					DBG_MSG(6, "Face vertex: %d, %d, %d", fv.g, fv.t, fv.n);
					f.push_back(fv);
					free(str);
				}
				face.push_back(f);
				
			} else if (strcmp(word, "mtllib") == 0) {
				// Materialy
				char *str, *fn;
				while (skip_space(file) && fscanf(file, "%a[^ \t\n\r]s", &str) == 1) {
					fn = make_filename(filename, str);
					DBG_MSG(6, "Material library: \"%s\"", fn);
					load_materials(filename, fn);
					free(fn);
					free(str);
				}

			} else if (strcmp(word, "usemtl") == 0) {
				// use material
				char *str;
				if (fscanf(file, " %a[^\n\r]s", &str) == 1) {
					usemtl = & material[str];
					free(str);
				}

			} else {
				// unknown key
				DBG_MSG(6, "Unknown key %s", word);
			}
			skip_to_eol(file);
			free(word);
		}
	}
	
	// close file
	fclose(file);
	return true;
}

bool GGLModel::load_materials(char *basefilename, char *filename) {
	FILE *file;
	char *newmtl_name = NULL;
	MATERIAL newmtl;
	newmtl.texture_loaded = false;

	if (!filename)
		return false;

	// open file
	if (!(file = fopen(filename, "rt"))) {
		perror(filename);
		return false;
	}

	// read & parse all lines
	while (!feof(file)) {
		char *word = NULL;
		if (fscanf(file, " %as", &word) == 1) {
			if (word[0] == '#') {
				// comment
				//DBG_MSG(6, "Skipped comment");
				
			} else if (strcmp(word, "newmtl") == 0) {
				if (newmtl_name) {
					material[newmtl_name] = newmtl;
					free(newmtl_name);
					newmtl_name = NULL;
					newmtl.have_texture = false;
				}
				
				fscanf(file, " %a[^\n\r]s", &newmtl_name);
					
			} else if (strcmp(word, "Ka") == 0) {
				// ambient color
				fscanf(file, " %f %f %f", &newmtl.ambient.r, &newmtl.ambient.g, &newmtl.ambient.b);
				DBG_MSG(6, "Loaded ambient: %.3f, %.3f, %.3f",
						newmtl.ambient.r, newmtl.ambient.g, newmtl.ambient.b);

			} else if (strcmp(word, "Kd") == 0) {
				// diffuse color
				fscanf(file, " %f %f %f", &newmtl.diffuse.r, &newmtl.diffuse.g, &newmtl.diffuse.b);
				DBG_MSG(6, "Loaded diffuse: %.3f, %.3f, %.3f",
						newmtl.diffuse.r, newmtl.diffuse.g, newmtl.diffuse.b);

			} else if (strcmp(word, "Ks") == 0) {
				// specular color
				fscanf(file, " %f %f %f", &newmtl.specular.r, &newmtl.specular.g, &newmtl.specular.b);
				DBG_MSG(6, "Loaded specular: %.3f, %.3f, %.3f",
						newmtl.specular.r, newmtl.specular.g, newmtl.specular.b);

			} else if (strcmp(word, "Ns") == 0) {
				// shininess
				fscanf(file, " %f", &newmtl.shininess);
				DBG_MSG(6, "Loaded shininess: %.3f", newmtl.shininess);

			} else if (strcmp(word, "map_Kd") == 0) {
				// texture file
				char *fn, *str;
				if (fscanf(file, " %a[^\n\r]s", &str) == 1) {
					fn = make_filename(basefilename, str);
					newmtl.texture = fn;
					newmtl.have_texture = true;
					DBG_MSG(6, "Texture: \"%s\"", newmtl.texture.c_str());
					free(fn);
					free(str);
				}
			} else {
				// unknown key
				DBG_MSG(6, "Unknown key \"%s\"", word);
			}
			skip_to_eol(file);
			free(word);
		}
	}
	if (newmtl_name) {
		material[newmtl_name] = newmtl;
		free(newmtl_name);
		newmtl_name = NULL;
	}
	
	// close file
	fclose(file);
	return true;
}

bool GGLModel::draw(void) {
	float real_scale = 1.0;
	float move_up = 0.0;
	float off_x = 0.0, off_y = 0.0, off_z = 0.0;
	MATERIAL *last_mtl = NULL;
	int face_type[] = { 0, 0, GL_LINES, GL_TRIANGLES, GL_QUADS, GL_POLYGON, 0 };

	/* test
	GLfloat ma[] = { 0.6, 0.6, 0.9};
	GLfloat md[] = { 0.1, 0.1, 0.3};
	GLfloat ms[] = { 0.0, 0.0, 0.0};
	GLfloat sh[] = { 0.3 };
	glColor3f((ma[0] + md[0])/2.0, (ma[1] + md[1])/2.0, (ma[2] + md[2])/2.0);
	glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ma);
	glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, md);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, ms);
	glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, sh);
	glutSolidTeapot(0.5);
	return true;
	// */
	
	if (size_forced) {
		float max = size_max[0] - size_min[0];
		for (int i = 1; i < 3; i++) {
			float d = size_max[i] - size_min[i];
			max = d > max ? d : max;
		}
		real_scale = size_scale / max;

		if(size_nooffset) {
			off_x = size_min[0] + (size_max[0] - size_min[0]) / 2.0;
			off_y = size_min[1] + (size_max[1] - size_min[1]) / 2.0;
			off_z = size_min[2] + (size_max[2] - size_min[2]) / 2.0;
			move_up = 0.5;
		}
	}
	
	for (FACE_ARRAY::iterator i = face.begin(); i != face.end(); i++) {
		// set texture
		FACE::iterator vb = i->begin();
		if (vb != i->end()) {
			if (vb->material->have_texture) {
				pngInfo info;
				if (!vb->material->texture_loaded) {
					DBG_MSG(4, "Loading texture \"%s\"", vb->material->texture.c_str());
					vb->material->texture_id = pngBind(vb->material->texture.c_str(),
							PNG_NOMIPMAP, PNG_ALPHA,
							&info, GL_CLAMP, GL_LINEAR, GL_LINEAR);
					vb->material->texture_loaded = (vb->material->texture_id != 0);
					if (vb->material->texture_loaded) {
						DBG_MSG(4, "Texture: Size=%i,%i Depth=%i Alpha=%i\n",
								info.Width, info.Height, info.Depth, info.Alpha);
					}
				}
				if (vb->material->texture_loaded) {
					DBG_MSG(6, "Using texture \"%s\" %d",
							vb->material->texture.c_str(), vb->material->texture_id);
					glEnable(GL_TEXTURE_2D);
					glBindTexture(GL_TEXTURE_2D, vb->material->texture_id);
				} else {
					DBG_MSG(1, "Texture loading failed.");
					vb->material->have_texture = false;
				}
			} else {
				DBG_MSG(6, "No texture");
				glDisable(GL_TEXTURE_2D);
			}
		}
		
		DBG_MSG(6, "Face vertexes: %d", i->size());
		glBegin(face_type[i->size()]);
		for (FACE::iterator v = i->begin(); v != i->end(); v++) {
			VERTEX *x = &geom_vertex[v->g - 1];

			if (last_mtl != v->material) {
				// set material
				GLfloat ma[] = { v->material->ambient.r,
					v->material->ambient.g, v->material->ambient.b };
				GLfloat md[] = { v->material->diffuse.r,
					v->material->diffuse.g, v->material->diffuse.b };
				GLfloat ms[] = { v->material->specular.r,
					v->material->specular.g, v->material->specular.b };
				GLfloat sh[] = { v->material->shininess };
				glColor3f((ma[0] + md[0])/2.0, (ma[1] + md[1])/2.0, (ma[2] + md[2])/2.0);
				glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, ma);
				glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, md);
				glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, ms);
				glMaterialfv(GL_FRONT_AND_BACK, GL_SHININESS, sh);
				
				
				last_mtl = v->material;

				DBG_MSG(6, "Face vertexes: %d", i->size());
				glBegin(face_type[i->size()]);
			}
		
			if (v->n != -1) {
				glNormal3f(norm_vertex[v->n - 1].x, norm_vertex[v->n - 1].y, norm_vertex[v->n - 1].z);
			}
			if (v->material->texture_loaded) {
				float x = texture_vertex[v->t - 1].x,
				      y = texture_vertex[v->t - 1].y;
				DBG_MSG(6, "Texture coord: %.3f, %.3f", x, y);
				glTexCoord2f(x, y);
			}
			DBG_MSG(6, "Vertex: %.3f, %.3f, %.3f", x->x, x->y, x->z);
			glVertex3f(	(x->x - off_x) * real_scale,
					(x->y - off_y) * real_scale + move_up,
					(x->z - off_z) * real_scale);
		}
		glEnd();
	}
	return true;
}

void GGLModel::skip_to_eol(FILE *file) {
	while (file && !feof(file) && getc(file) != '\n')
		;
}

bool GGLModel::skip_space(FILE *file) {
	int c;
	while (file && !feof(file)) {
		c = getc(file);
		if (c != ' ' && c != '\t') {
			ungetc(c, file);
			return true;
		}
	}
	return false;
}

char *GGLModel::make_filename(char *dir_file, char *new_file) {
	char *path = NULL, *slash = NULL, *new_dirfile = NULL;

	if (!dir_file || !new_file)
		return NULL;

	DBG_MSG(7, "dir_file = \"%s\", new_file = \"%s\"", dir_file, new_file);
	
	path = strdup(dir_file);
	if ((slash = strrchr(path, '/'))) {
		*slash = 0;
		new_dirfile = (char*) malloc(strlen(path) + strlen(new_file) + 2);
		strcpy(new_dirfile, path);
		strcat(new_dirfile, "/");
		strcat(new_dirfile, new_file);
	} else {
		new_dirfile = strdup(new_file);
	}
	free(path);
	DBG_MSG(7, "new_dirfile = \"%s\"", new_dirfile);
	return new_dirfile;
}

