/*
 * game_object.cc
 * 
 * Copyright (C) 2005  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * 
 */

#include "game_object.h"

#define DEBUG 0
#include "debug.h"

GObject::GObject(void) {
	name = "object";
	goid = GOID_UNDEFINED;
	height = 0;
	width = 0;
	depth = 0;
	death = false;
	add_property(GOP_POS_X, &pos_x);
	add_property(GOP_POS_Y, &pos_y);
	add_property(GOP_POS_Z, &pos_z);
	add_property(GOP_HEIGHT, &height, true);
	add_property(GOP_WIDTH, &width, true);
	add_property(GOP_DEPTH, &depth, true);
}

GObject::~GObject(void) {
	for (GObjectList::iterator i = child.begin(); i != child.end(); i++) {
		delete (*i);
		child.erase(i);
	}
	for (GOPList::iterator i = gop_list.begin(); i != gop_list.end(); i++) {
		gop_list.erase(i);
	}
}
		
void GObject::set_position(int pos_x, int pos_y, int pos_z) {
	this->pos_x = pos_x;
	this->pos_y = pos_y;
	this->pos_z = pos_z;
}

void GObject::step() {
	update_get_child_at();
}

void GObject::update_get_child_at(void) {
	int key;
	child_map.clear();
	
	for (GObjectList::iterator i = child.begin(); i != child.end(); i++) {
		if (!(*i)->parent)
			key = 0;
		else
			key = ((*i)->pos_x + (*i)->pos_y * width) * depth + (*i)->pos_z;
		child_map.insert(std::make_pair(key, *i));
	}
}
		
GObjectList GObject::get_child_at(int x, int y) {
	GObjectList ret;
	
	for (GObjectMap::iterator i = child_map.lower_bound((x + y * width) * depth);
			i != child_map.upper_bound((x + y * width) * depth + 1); i++) 
	{
		ret.push_back(i->second);
	}
	
	return ret;
}

GObject *GObject::get_child_by_goid(int goid) {
	for (GObjectList::iterator i = child.begin(); i != child.end(); i++) {
		if ((*i)->goid == goid)
			return (*i);
	}
	return NULL;
}

void GObject::add_property(const int prop_id, GOPType *var, bool readonly) {
	if (var) {
		DBG_MSG(9, "GObject::add_property: adding prop_id %d, readonly = %s",
				prop_id, readonly ? "true":"false");
		gop_list.insert(std::make_pair(prop_id, var));
	}
}

bool GObject::set_property(const int prop_id, GOPType val) {
	DBG_MSG(9, "GObject::set_property: setting prop_id %d to value %d", prop_id, (int)val);
	GOPList::iterator prop = gop_list.find(prop_id);
	if(prop != gop_list.end()) {
		*(prop->second) = val;
		return true;
	} else {
		return false;
	}
}

bool GObject::add_to_property(const int prop_id, GOPType val) {
	DBG_MSG(9, "GObject::add_to_property: adding to prop_id %d value %d", prop_id, (int)val);
	GOPList::iterator prop = gop_list.find(prop_id);
	if(prop != gop_list.end()) {
		*(prop->second) += val;
		return true;
	} else {
		return false;
	}
}

GOPType GObject::get_property(const int prop_id) {
	DBG_MSG(9, "GObject::set_property: reading prop_id %d", prop_id);
	GOPList::iterator prop = gop_list.find(prop_id);
	if(prop != gop_list.end()) {
		DBG_MSG(9, "GObject::set_property: prop_id %d has value %d", prop_id, (int)*(prop->second));
		return *(prop->second);
	} else {
		return 0;
	}
}
		
/****************************************************************************/

