/*
 * screen-curses.cc
 * 
 * Copyright (C) 2005  Josef Kufner <jk@myserver.cz>
 * 
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <ncurses.h>
#include <signal.h>
#include <strings.h>

#include "screen-curses.h"
#include "game_world.h"
#include "game_object.h"
#include "sample_msg.h"
#include "sound_const.h"

#define CLASS "Screen"
#define DEBUG 0
#include "debug.h"

#define ATTR_ST			(COLOR_PAIR(1))
#define ATTR_ST_LIGHT		(COLOR_PAIR(2) | A_BOLD)
#define ATTR_ST_DARK		(COLOR_PAIR(3) | A_BOLD)
#define ATTR_BORDER		(COLOR_PAIR(1))
#define ATTR_FREQ		(COLOR_PAIR(4))
#define ATTR_FREQ_H		(COLOR_PAIR(5) | A_BOLD)
#define ATTR_FREQ_RULE		(COLOR_PAIR(6))
#define ATTR_FREQ_RULE_INT	(COLOR_PAIR(7) | A_BOLD)
#define ATTR_FREQ_RULE_INT_H	(COLOR_PAIR(8) | A_BOLD)

/*****************************************************************************
 * Inicializace obrazovky
 */
ScreenCurses::ScreenCurses (ScreenFifo *fifo) : Thread("ScreenCurses", true) {
	DBG_MSG(1, "Initializing...");

	this->fifo = fifo;
	this->fifo->target_ref();

	/* ncurses init */
	initscr();
	noecho();
	cbreak();
	crmode();
	keypad(stdscr, 1);
	if (has_colors()) {
		DBG_MSG(6, "Initializing colors...");
		start_color();
		init_pair(1, COLOR_WHITE,	COLOR_BLACK);
		init_pair(2, COLOR_YELLOW,	COLOR_BLACK);
		init_pair(3, COLOR_BLACK,	COLOR_BLACK);
		init_pair(4, COLOR_CYAN,	COLOR_BLACK);
		init_pair(5, COLOR_WHITE,	COLOR_BLACK);
		init_pair(6, COLOR_WHITE,	COLOR_BLACK);
		init_pair(7, COLOR_WHITE,	COLOR_BLUE);
		init_pair(8, COLOR_WHITE,	COLOR_GREEN);

		/* barvy pro objekty */
		init_pair(10, COLOR_BLACK,	COLOR_BLACK);
		init_pair(11, COLOR_BLUE,	COLOR_BLACK);
		init_pair(12, COLOR_GREEN,	COLOR_BLACK);
		init_pair(13, COLOR_RED,	COLOR_BLACK);
		init_pair(14, COLOR_CYAN,	COLOR_BLACK);
		init_pair(15, COLOR_MAGENTA,	COLOR_BLACK);
		init_pair(16, COLOR_YELLOW,	COLOR_BLACK);
		init_pair(17, COLOR_BLACK,	COLOR_BLACK);
		init_pair(18, COLOR_BLACK,	COLOR_BLACK); // + A_BOLD
		init_pair(19, COLOR_BLUE,	COLOR_BLACK); //    :
		init_pair(20, COLOR_GREEN,	COLOR_BLACK); //    .
		init_pair(21, COLOR_RED,	COLOR_BLACK);
		init_pair(22, COLOR_CYAN,	COLOR_BLACK);
		init_pair(23, COLOR_MAGENTA,	COLOR_BLACK);
		init_pair(24, COLOR_YELLOW,	COLOR_BLACK);
		init_pair(25, COLOR_WHITE,	COLOR_BLACK);
	} else {
		fprintf(stderr, "Cannot initialize colors!\n");
	}
	
	resized = 0;
	draw_all();

	DBG_MSG(2, "Done.");
	thread_launch();
}


/*****************************************************************************
 * Vyitn obrazovky
 */
ScreenCurses::~ScreenCurses(void) {
	DBG_MSG(1, "Cleaning up...");
	
	/* ncurses */
	endwin();

	DBG_MSG(2, "Done.");
}

/*****************************************************************************
 * Thread main
 */
void *ScreenCurses::main (void) {
	const float freq_zoom = 4.0;
	/*
	screen->render_world(world);
	screen->draw_all();
	*/
	DBG_MSG(3, "");
	ScreenMsg *msg;
	while (fifo->pop(&msg)) {
		DBG_MSG(5, "Message received.");
		
		// null
		if (!msg) {
			DBG_MSG(5, "Empty message.");
			continue;
		}

		// World size
		if (msg->have_world_size()) {
			DBG_MSG(5, "World size");
			msg->get_world_size(&world_width, &world_height);
		}

		// GSymbol
		if (msg->have_symbols()) {
			DBG_MSG(5, "GSymbol");
			draw_world(msg);
		}

		// ScreenFreqRule
		if (msg->have_freq_rule()) {
			DBG_MSG(5, "ScreenFreqRule");
			freq_rule = msg->get_freq_rule();
		}

		// SampleMsg
		if (msg->have_sample_msg()) {
			DBG_MSG(5, "SampleMsg");
			SampleMsg *sample_msg = msg->get_sample_msg();
			render_freq(sample_msg->freq, (int)(sample_msg->size / freq_zoom),
					sample_msg->freq_minval, sample_msg->intensity, MIN_INTENSITY);
			draw_freq();
		}
		
		// Status line
		if (msg->have_status()) {
			int k_key, k_score, z_key, z_score;
			msg->get_status(&k_key, &k_score, &z_key, &z_score);
			draw_status(k_key, k_score, z_key, z_score);
		}

		// done
		flush();
		msg->unref();
	}
	fifo->target_unref();
	
	DBG_MSG(3, "Bye.");
	return NULL;
}

/*****************************************************************************
 * Vykresleni celeho sveta -- neumi to scrollovani, ale aspon to vycentruje
 * svet na obrazovce
 */
void ScreenCurses::render_world(GameWorld *world) {

	DBG_MSG(2, "");
	GObject *object = NULL;
	GSymbol *new_symbol = NULL;

	DBG_MSG(2, "Rendering world (width = %d, height = %d)...", world_width, world_height);

	for (SymbolBuffer::iterator i = symbol.begin(); i != symbol.end(); i++) {
		if (*i)
			delete *i;
	}
	symbol.clear();
	symbol.resize(world_height * world_width, 0);

	world->get_child_rewind();
	while ((object = world->get_child())) {
		DBG_MSG(9, "Drawing symbol for object \"%s\"", object->get_name());
		new_symbol = object->draw();
		DBG_MSG(9, "Found object (symbol is '%c').", new_symbol->symbol);
		if (new_symbol->x >= 0 && new_symbol->x < world_width
				&& new_symbol->y >= 0 && new_symbol->y < world_height
				&& new_symbol->z >= 0 && new_symbol->z <= 2)
		{
			int symbol_i = new_symbol->x + new_symbol->y * world_width;
			if (symbol[symbol_i] && new_symbol->z > symbol[symbol_i]->z) {
				delete symbol[symbol_i];
				symbol[symbol_i] = 0;
			}
			if (!symbol[symbol_i]) {
				symbol[symbol_i] = new_symbol;
				DBG_MSG(9, "Object stored (symbol is '%c').",
						symbol[symbol_i]->symbol);
			} else {
				delete new_symbol;
			}
			DBG_MSG(9, "Symbol stored");
		} else {
			DBG_MSG(9, "Object is out, discarded.");
			delete new_symbol;
		}
	}

	DBG_MSG(2, "Done.");
}


/*****************************************************************************
 * Renderovani frekvencniho spektra do pole intu podle velikosti obrazovky
 *
 * Prevede pole floatu o velke sirce na pole intu o sirce obrazovky.
 *
 * Pri zmene velikosti obrazovky spektrum zmizi a vykresli se az po dalsim
 * renderovani
 *
 * prepocet:
 * 	1. sloucit frekvence do sloupcu (pokud je obrazovka sirsi nez spektrum,
 *	   tak se jen kopiruje)
 *  	   - zaklad: 1 sloupec grafu = >1 sloupec spektra
 *  	   
 * 	2. upravit vysku sloupcu
 * 	   - najit maximum == nejvyssi sloupec
 * 	   - prepocitat screen_height - world_height - 5 == nejvyssi sloupec
 *
 * max intensity = 35'000'000
 */
void ScreenCurses::render_freq(int *freq, int size, int min_colorized, int intensity, int intensity_min) {
	int do_zoom = screen_width - 2 < size;
	int bar_size = do_zoom ? screen_width - 2 : size;
	int bar[bar_size];
	int max = 0;
	DBG_MSG(2, "size = %d, min_colorized = %d; do_zoom = %d",
			size, min_colorized, do_zoom);
	
	bzero(bar, bar_size * sizeof(int));

	
	/* slouceni frekvenci */
	if (do_zoom) {
		int f = 0, b = 0;
		int f_step = (int)((float)size / (float)bar_size);
		/* kaslu na zbytecek na konci */
		while (b < bar_size && f < size) {
			if (f % f_step == f_step - 1) {
				max = bar[b] > max ? bar[b] : max;
				b++;
			}
			bar[b] += freq[f++];
		}
		hscale = (float)b / (float)f;
	} else /* screen_width >= size */ {
		for (int i = 0; i < size; i++) {
			bar[i] = freq[i];
			max = bar[i] > max ? bar[i] : max;
		}
		memcpy(bar, freq, sizeof(int) * size);
		hscale = 1.0;
	}

	/* uprava vysek */
	bar_height = screen_height - world_height - 5;
	bar_width = bar_size;
	freq_bar.resize(bar_size);
	float mul = (float)bar_height / (float)max;
	for (int i = 0; i < bar_size; i++)
		freq_bar[i] = (int)(bar[i] * mul);
	freq_bar_max = max;
	freq_bar_min_colorized = (int)(mul * min_colorized);

	/* uprava intensity */
	this->intensity = (int)(((float)intensity / (float)(intensity_min * 8)) * bar_size);
	this->intensity_min = (int)(((float)intensity_min / (float)(intensity_min * 8)) * bar_size);
	
	DBG_MSG(4, "max = %d, min_colorized = %d",
			freq_bar_max, freq_bar_min_colorized);
}


/*****************************************************************************
 * Prekresleni obrazovky
 */
void ScreenCurses::redraw(void) {
	DBG_MSG(2, "Redrawing screen...");
	endwin();
	refresh();
	draw_all();
	touchwin(stdscr);
	refresh();
	DBG_MSG(2, "Redrawing screen... Done.");
}


/*****************************************************************************
 * Ulozeni flagu, ze se ma obrazovka refreshnout dukladne
 */
void ScreenCurses::resize_notify(void) {
	resized = 1;
}


/*****************************************************************************
 * Vykresli vsechno
 */
void ScreenCurses::draw_all(void) {
	if (resized) {
		endwin();
		refresh();
		resized = 0;
	}
	erase();
	getmaxyx(stdscr, screen_height, screen_width);
	//draw_status();
	//draw_world();
	//draw_freq();
	flush();
}



/*****************************************************************************
 * Vykresleni hraciho sveta
 */
void ScreenCurses::draw_world(ScreenMsg *msg) {
	GSymbol *sym = NULL;
	char *depth_buffer = NULL;

	if (resized) {
		draw_all();
//		return;
	}

	DBG_MSG(2, "Drawing world (screen_width = %d, screen_height = %d)...",
			screen_width, screen_height);
	
	//int voff = ((screen_height - world_height - 2) / 2) + 2;
	int voff = 3;
	int hoff = ((screen_width - world_width) / 2);

	/* ramecek */
	attrset(ATTR_BORDER);
	mvaddch(voff - 1, hoff - 1, ACS_ULCORNER);
	mvaddch(voff - 1, hoff + world_width, ACS_URCORNER);
	mvaddch(voff + world_height, hoff - 1, ACS_LLCORNER);
	mvaddch(voff + world_height, hoff + world_width, ACS_LRCORNER);
	
	for (int col = 0; col < world_width; col++)
		mvaddch(voff - 1, hoff + col, ACS_HLINE);
	
	for (int col = 0; col < world_width; col++)
		mvaddch(voff + world_height, hoff + col, ACS_HLINE);

	for (int row = 0; row < world_height; row++)
		mvaddch(voff + row, hoff - 1, ACS_VLINE);
	
	for (int row = 0; row < world_height; row++)
		mvaddch(voff + row, hoff + world_width, ACS_VLINE);


	// init depth buffer
	if(!(depth_buffer = (char *)malloc(world_width * world_height))) {
		perror("depth_buffer");
		exit(-1);
	}
	bzero(depth_buffer, world_width * world_height);
	
	/* svet */
	msg->get_symbol_rewind();
	while ((sym = msg->get_symbol())) {
		if (sym->z + 1 > depth_buffer[sym->y * world_width + sym->x]) {
			attrset(COLOR_PAIR(sym->color + 10) | (sym->color > 7 ? A_BOLD : 0));
			mvaddch(voff + sym->y, hoff + sym->x, sym->symbol ? sym->symbol : ' ');
			depth_buffer[sym->y * world_width + sym->x] = sym->z + 1;
		}
	}

	free(depth_buffer);
	DBG_MSG(2, "Done.");
}


/*****************************************************************************
 * Vykresleni cudliku
 */
void ScreenCurses::draw_status(int k_key, int k_score, int z_key, int z_score) {
	if (resized) {
		draw_all();
		return;
	}

	DBG_MSG(6, "drawing...");

	/*zahradnik */
	move(0, 2);
	attrset(ATTR_ST);
	printw("Zahradnik: ");
	
	attrset(z_key & 2 ? ATTR_ST_LIGHT : ATTR_ST_DARK);	
	printw(" <-");
	
	attrset(ATTR_ST);
	printw(" %4d ", z_score);
	
	attrset(z_key & 1 ? ATTR_ST_LIGHT : ATTR_ST_DARK);	
	printw(" ->");

	/* Krtek */
	move(0, screen_width - 20);
	attrset(ATTR_ST);
	printw("Krtek: ");
	
	attrset(k_key & 2 ? ATTR_ST_LIGHT : ATTR_ST_DARK);	
	printw(" <-");
	
	attrset(ATTR_ST);
	printw(" %4d ", k_score);
	
	attrset(k_key & 1 ? ATTR_ST_LIGHT : ATTR_ST_DARK);	
	printw(" ->");

	move(1, 0);
	attrset(ATTR_ST);
	for (int i = 0; i < screen_width; i++)
		insch(ACS_HLINE);

}


/*****************************************************************************
 * Vykresleni frekv. spektra
 */
void ScreenCurses::draw_freq(void) {
	float from, to;
	char label;
	
	DBG_MSG(6, "drawing...");
	if (resized) {
		draw_all();
		return;
	}

	int voff = screen_height - bar_height;
	unsigned int i;

	if (resized)
		draw_all();

	if (voff >= screen_height)
		return;
	
	for (i = 0; i < freq_bar.size(); i++) {
		attrset(freq_bar[i] > freq_bar_min_colorized ? ATTR_FREQ_H : ATTR_FREQ);
		for (int r = 0; r < bar_height - 1; r++) {
			mvaddch(voff + r, i + 1, freq_bar[i] > bar_height - r ? '#' : ' ');
		}
	}
	
	/* pravitko */
	attrset(ATTR_FREQ_RULE_INT);
	for (i = 0; i < freq_bar.size() - 1; i++) {
		if (i == (unsigned)intensity_min) {
			attrset(ATTR_FREQ_RULE_INT_H);
		}
		if (i > (unsigned)intensity) {
			attrset(ATTR_FREQ_RULE);
		}
		freq_rule->get_rewind();
		while (freq_rule->get(&from, &to, &label)) {
			if (i > from * hscale && i < to * hscale) {
				mvaddch(screen_height - 1, i + 2, label);
				break;
			} else {
				mvaddch(screen_height - 1, i + 2, ' ');
			}
		}
	}
	attrset(ATTR_FREQ_RULE);
	mvaddch(screen_height - 1, 1, ACS_VLINE);
	mvaddch(screen_height - 1, freq_bar.size(), ACS_VLINE);
	DBG_MSG(6, "done");
}


/*****************************************************************************
 * Jen opravdu vykresleni toho co je pripraveno
 */
void ScreenCurses::flush(void) {
	if (resized)
		draw_all();
	move(screen_height - 1, 0);
	refresh();
}

#undef CLASS

